#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <string.h>

void initialize( float* x, int n )
{
  int n2 = n+2;
  for (int i=0; i<n2; ++i)
    for (int j=0; j<n2; ++j)
      x[i*n2 + j] = random() / (float) RAND_MAX;
}

__global__ void smooth( float* y, float* x, int n, float a, float b, float c )                                             {                                                                                                                               
    int n2 = n+2, i, j;                                                                                                        
                                                                                                                               
    i = blockDim.x * blockIdx.x + threadIdx.x + 1;  // add one to start from index 1                                            
    j = blockDim.y * blockIdx.y + threadIdx.y + 1;                                                                                 
    y[i*n2 + j] = a * (x[(i-1)*n2 + (j-1)] + x[(i-1)*n2 + (j+1)] + x[(i+1)*n2 + (j-1)] + x[(i+1)*n2 + (j+1)])+b*(x[i*n2+(j-1)] + x[i*n2+(j+1)] + x[(i-1)*n2 + j] + x[(i+1)*n2 + j]) + c * x[i*n2 +j];                                                                  
}

void count( float* x, int n, float t, int &nbx )
{
  nbx = 0;
  int n2 = n+2;
  for (int i=1; i <= n; ++i)
    for (int j=1; j <= n; ++j)
      if (x[i*n2 + j] < t)
        ++nbx;
}

class Timer
{
public:
  Timer(): n(0) { }
  void start(std::string label) 
  { 
    if (n < 20) 
      { labels[n] = label; times[2*n] = clock(); }
    else { std::cerr << "No more timers, " << label 
      << " will not be timed." << std::endl; }
  }

  void stop() { times[2*n+1] = clock(); n++;}
  void reset() { n=0; }
  void print();
private:
  std::string labels[20];
  float times[40];
  int n;
};


void Timer::print()
{
    std::cout << std::endl;
    std::cout << "Action           ::   time/s     Time resolution = " << 1.f/(float)CLOCKS_PER_SEC << std::endl;
    std::cout << "------" << std::endl;
    for (int i=0; i < n; ++i)
        std::cout << labels[i] << " :: " << (times[2*i+1] - times[2*i+0])/(float)CLOCKS_PER_SEC << std::endl;
}

int main()
{

int n,nbx,nby;

float a,b,c,t,*x,*y;

/* timer of class Timer */
Timer timer, timer2;

/* n - number of elements in one direction */
    n = 1 << 14;  /* 2^14 == 16384 */

/* a,b,c - smoothing constants */
    a = 0.05;
    b = 0.1;
    c = 0.4;

/* t - threshold */
    t = 0.5;

float *x_d, *y_d;

/* allocate device input array x_d */                                                                     
timer.start("GPU: Alloc-X_D  ");                                                                          
cudaMalloc((void**) &x_d, (n+2)*(n+2)*sizeof(float));                                                     
timer.stop();                                                                                             
/* allocate device output array y_d */                                                                    
timer.start("GPU: Alloc-Y_D  ");                                                                          
cudaMalloc((void**) &y_d, (n+2)*(n+2)*sizeof(float));                                                     
timer.stop(); 


/* allocate x */
    x = new float[(n+2)*(n+2)];

/* allocate y */
    y = new float[(n+2)*(n+2)];

 /* initialize x */
initialize(x, n);
    
/* smooth x into y */
//smooth(y, x, n, a, b, c);

/* copy x_h to x_d on GPU */                                                                              
timer.start("GPU: Copy-in    ");                                                                          
cudaMemcpy(x_d, x, (n+2)*(n+2)*sizeof(float), cudaMemcpyHostToDevice);                                    
timer.stop();                                                                                             
/* launch smooth() on GPU */                                                                              
int nBlocks = 16;                                                                                         
timer.start("GPU: Smooth     ");                                                                          
smooth<<<n/nBlocks,nBlocks>>>(y_d, x_d, n, a, b, c);                                                 
cudaDeviceSynchronize();                                                                                  
timer.stop();                                                                                             
/* copy y_d to y_h on CPU */                                                                              
timer.start("GPU: Copy-out   ");                                                                          
cudaMemcpy(y, y_d, (n+2)*(n+2)*sizeof(float), cudaMemcpyDeviceToHost);                                  
timer.stop();                                                                                             
/* count elements in second array */
timer2.start("CPU: Count-XY");
  timer.start("CPU: Count-X");
  count(x, n, t, nbx);
  timer.stop(); 
  timer.start("CPU: Count-Y");
  count(y, n, t, nby);
  timer.stop();
timer2.stop();

std::cout << std::endl;
std::cout << "Summary" << std::endl;
std::cout << "-------" << std::endl;
std::cout << "Number of elements in a row/column         :: " << n+2 << std::endl;
std::cout << "Number of inner elements in a row/column   :: " << n << std::endl;
std::cout << "Total number of elements                   :: " << (n+2)*(n+2) << std::endl;
std::cout << "Total number of inner elements             :: " << n*n << std::endl;
std::cout << "Memory (GB) used per array                 :: " << (n+2)*(n+2)*sizeof(float) / (float)(1024*1024*1024) << std::endl;;
std::cout << "Threshold                                  :: " << t << std::endl;
std::cout << "Smoothing constants (a, b, c)              :: " << a << " " << b << " " << c << std::endl;
std::cout << "Number   of elements below threshold (X)   :: " << nbx << std::endl;
std::cout << "Fraction of elements below threshold       :: " << nbx / (float)(n*n) << std::endl;
std::cout << "Number   of elements below threshold (Y)   :: " << nby << std::endl;
std::cout << "Fraction of elements below threshold       :: " << nby / (float)(n*n) << std::endl;

timer.print();
timer2.print();

return(0);
}
